package backup.client;

import java.security.MessageDigest;
import java.text.SimpleDateFormat;
import java.util.*;
import java.io.*;

/**
 * Main client engine.
 * 
 * @author <a href="mailto:mauset@idi.ntnu.no">Tore Mauset</a>
 */
public class BackupClient extends Thread {
    private UserInterface ui;

    private RdiffWrapper rdiff;

    // Encrypted or unencrypted connection
    // private SSLConnection conn;
    private Connection conn;
    private ArrayList<String> targets;
    private ArrayList<String> ignores;
    private String host;
    private int port;
    private String hostname;
    private String user;
    private String password;
    private Boolean autoBackup;
    private String autoBackupTime;
    private Boolean autoLogin;
    private String rdiffCommand;

    private boolean running;
    private String osName;
    private boolean backupIsRunning = false;
    private Config config;
    private boolean loggedIn = false;
    private String nextChallenge;
    private final String configFilename = "config.dat";

    /**
     * Creates a new <code>BackupClient</code>.
     * 
     * @param args command line arguments
     */
    public BackupClient(String[] args) {

        if (args.length > 0 && args[0].equals("-g")) {
            ui = new Gui(this);
        }
        else {
            ui = new Cli(this);
        }

        osName = System.getProperty("os.name");

        loadConfig();

        rdiff = new RdiffWrapper(ui);

        running = true;

        if (autoLogin) {
            login();
        }
    }

    /**
     * Main.
     * 
     * @param args command line arguments
     */
    public static void main(String[] args) {
        BackupClient bc = new BackupClient(args);
        bc.run();
    }

    /**
     * Start main thread. This thread keeps track of tasks performed at regular
     * intervals.
     */
    public void run() {
	SimpleDateFormat formatter = new SimpleDateFormat("HH:mm");
	String prev = formatter.format(new Date());
	int mins = 0;

	try {
	    while (running) {
		Thread.sleep(1000);
		String now = formatter.format(new Date());
		if (!now.equals(prev)) {
		    prev = now;
		    mins++;

		    // Send keep-alive signal to server every minute
		    if (isLoggedIn()) {
			keepAlive();
		    }

		    // Start scheduled backup
		    if (autoBackup && autoBackupTime.equals(now)) {
			startBackup();
		    }

		    // Reset counter every hour
		    if (mins == 60) {
			mins = 0;
		    }
		}
            }
	}
	catch (InterruptedException e) {
	    e.printStackTrace();
	}
    }

    /**
     * Tell the <code>BackupClient</code> to shut down.
     */
    public void exit() {
        if (ui != null) {
            ui.status("Client shutting down.");
        }
        if (isLoggedIn()) {
            logout();
        }
        running = false;
    }

    /**
     * Log in to the server. Requests a challenge from the server and
     * responds to it.
     */
    public void login() {
        if (loggedIn) {
            return;
        }
        ui.status("Logging in...");

        // Encrypted or unencrypted connection
        // conn = new SSLConnection(this, host, port);
        conn = new Connection(this, host, port);

        Vector<String> query = new Vector<String>();
        query.add(user);

        // Start login sequence and request a challenge from the server
        conn.send("startLogin", query);

        query.clear();
        query.add(hash(hash(password) + nextChallenge));
        query.add(user);

        // Reply to the server with the response
        String success = (String)conn.send("finishLogin", query).get("success");

        if (success.equals("true")) {
            loggedIn = true;
            ui.status("Logged in.");
        }
        else {
            loggedIn = false;
            ui.warn("Login failed.");
        }
    }

    /**
     * Log out from the server.
     */
    public void logout() {
        if (!loggedIn) {
            return;
        }
        Vector<String> query = new Vector<String>();
        query.add(hash(hash(password) + nextChallenge));
        query.add(user);
        conn.send("logout", query);
        loggedIn = false;
        ui.status("Logged out.");
    }

    /**
     * Starts backup process.
     */
    public void startBackup() {
        if (backupIsRunning) {
            ui.warn("Backup already running, aborting.");
            return;
        }

        if (targets.isEmpty()) {
            ui.warn("No targets for backup, aborting.");
            return;
        }

        if (!loggedIn) {
            ui.warn("Not logged in, aborting.");
            return;
        }

        ui.status("Starting backup...");
        backupIsRunning = true;

        Vector<String> query = new Vector<String>();
        query.add(hash(hash(password) + nextChallenge));
        query.add(user);

        // Make sure target directory on the server is clean
        conn.send("checkBackup", query);

        ArrayList cmds = new ArrayList();
        if (osName.startsWith("Linux") || osName.startsWith("Mac")) {
            cmds = buildBackupCmdUnix();
        }
        else if (osName.startsWith("Windows")) {
            cmds = buildBackupCmdWin();
        }

        Boolean success = true;

        for (int i = 0; i < cmds.size(); i++) {
            String[] cmd = (String[])cmds.get(i);

            // Tell Rdiffwrapper to start and break out of loop if it fails.
            if (!rdiff.start(cmd)) {
                success = false;
                break;
            }
        }

        // If everything is ok, tell server to update database
        if (success) {
            query.clear();
            query.add(hash(hash(password) + nextChallenge));
            query.add(user);

            // Update the server's file database
            conn.send("updateDB", query);
            ui.status("Backup finished.");
        }
        else {
            ui.status("Backup failed.");
        }

        backupIsRunning = false;
    }

    /**
     * Starts restore process.
     * 
     * @param timestamp timestamp of target to restore, according to
     *            rdiff-backup standards
     * @param target target to restore
     * @param force if true, rdiff-backup will delete existing target before
     *            restoring
     */
    public void startRestore(String timestamp, String target, Boolean force) {
        if (backupIsRunning) {
            ui.warn("Backup already running, aborting.");
            return;
        }

        if (!loggedIn) {
            ui.warn("Not logged in, aborting.");
            return;
        }

        ui.status("Starting restore...");
        backupIsRunning = true;

        // In case of invalid target from the cli
        Hashtable result = listFiles(target);
        if (!result.isEmpty()) {
            String[] cmd = { "" };
            if (osName.startsWith("Linux") || osName.startsWith("Mac")) {
                cmd = buildRestoreCmdUnix(timestamp, target, force);
            }
            else if (osName.startsWith("Windows")) {
                cmd = buildRestoreCmdWin(timestamp, target, force);
            }

            if (rdiff.start(cmd)) {
                ui.status("Restore finished.");
            }
            else {
                ui.status("Restore failed.");
            }
        }
        else {
            ui.warn("Unknown target.");
        }

        backupIsRunning = false;
    }

    /**
     * Add new target for backup.
     * 
     * @param target new target
     */
    public void addTarget(String target) {
        if (targets.indexOf(target) == -1) {
            targets.add(target);
        }
    }

    /**
     * Remove existing target for backup.
     * 
     * @param target existing target
     */
    public void removeTarget(String target) {
        int i = targets.indexOf(target);
        if (i != -1) {
            targets.remove(i);
        }
    }

    /**
     * Return current targets.
     * 
     * @return current targets
     */
    public ArrayList listTargets() {
        return targets;
    }

    /**
     * Add a new ignore from backup.
     * 
     * @param ignore new ignore
     */
    public void addIgnore(String ignore) {
        if (ignores.indexOf(ignore) == -1) {
            ignores.add(ignore);
        }
    }

    /**
     * Remove an existing ignore.
     * 
     * @param ignore existing ignore
     */
    public void removeIgnore(String ignore) {
        int i = ignores.indexOf(ignore);
        if (i != -1) {
            ignores.remove(i);
        }
    }

    /**
     * Return current ignores.
     * 
     * @return current ignores
     */
    public ArrayList listIgnores() {
        return ignores;
    }

    /**
     * Request a list of files matching search term from BackupServer.
     * 
     * @param search search term
     * @return list of files
     */
    public Hashtable listFiles(String search) {
        if (!loggedIn) {
            ui.warn("Not logged in, aborting search.");
            return null;
        }

        Vector<String> query = new Vector<String>();
        query.add(hash(hash(password) + nextChallenge));
        query.add(user);
        query.add(search.replace("*", "%"));

        return conn.send("listFiles", query);
    }

    /**
     * Are we logged in on the server?
     * 
     * @return logged in
     */
    public boolean isLoggedIn() {
        return loggedIn;
    }

    /**
     * Get username.
     * 
     * @return username
     */
    public String getUser() {
        return user;
    }

    /**
     * Set username.
     * 
     * @param user username
     */
    public void setUser(String user) {
        this.user = user;
    }

    /**
     * Set password.
     * 
     * @param password password
     */
    public void setPassword(String password) {
        this.password = password;
    }

    /**
     * Get local hostname.
     * 
     * @return hostname
     */
    public String getHostname() {
        return hostname;
    }

    /**
     * Set local hostname.
     * 
     * @param hostname hostname
     */
    public void setHostname(String hostname) {
        this.hostname = hostname;
    }

    /**
     * Get server address.
     * 
     * @return server address
     */
    public String getHost() {
        return host;
    }

    /**
     * Set server address.
     * 
     * @param host server address
     */
    public void setHost(String host) {
        this.host = host;
    }

    /**
     * Get server port.
     * 
     * @return server port
     */
    public String getPort() {
        return String.valueOf(port);
    }

    /**
     * Set server port.
     * 
     * @param port server port
     */
    public void setPort(String port) {
        this.port = Integer.parseInt(port);
    }

    /**
     * Get automatic backup on or off
     * 
     * @return automatic backup
     */
    public Boolean getAutoBackup() {
        return autoBackup;
    }

    /**
     * Set automatic backup on or off
     * 
     * @param autoBackup automatic backup
     */
    public void setAutoBackup(Boolean autoBackup) {
        this.autoBackup = autoBackup;
    }

    /**
     * Get time of day for automatic backup in 24 hour format, e.g. "23:30".
     * 
     * @return time for automatic backup
     */
    public String getAutoBackupTime() {
        return autoBackupTime;
    }

    /**
     * Set time of day for automatic backup in 24 hour format, e.g. "23:30".
     * 
     * @param autoBackupTime time for automatic backup
     */
    public void setAutoBackupTime(String autoBackupTime) {
        this.autoBackupTime = autoBackupTime;
    }

    /**
     * Get automatic login on startup on or off.
     * 
     * @return automatic login
     */
    public Boolean getAutoLogin() {
        return autoLogin;
    }

    /**
     * Set automatic login on startup on or off.
     * 
     * @param autoLogin automatic login
     */
    public void setAutoLogin(Boolean autoLogin) {
        this.autoLogin = autoLogin;
    }

    /**
     * Get path to rdiff-backup executable
     * 
     * @return rdiff-backup command
     */
    public String getRdiffCommand() {
        return rdiffCommand;
    }

    /**
     * Set path to rdiff-backup executable
     * 
     * @param rdiffCommand rdiff-backup executable
     */
    public void setRdiffCommand(String rdiffCommand) {
        this.rdiffCommand = rdiffCommand;
    }

    /**
     * Pass a warning to the user interface.
     * 
     * @param s warning text
     */
    public void warn(String s) {
        ui.warn(s);
    }

    /**
     * Records a new challenge for authentification of next message.
     * 
     * @param challenge the new challenge
     */
    public void setChallenge(String challenge) {
        nextChallenge = challenge;
    }

    /**
     * Saves the configuration to disk.
     */
    public void saveConfig() {
        try {
            config.targets = targets;
            config.ignores = ignores;
            config.host = host;
            config.port = port;
            config.hostname = hostname;
            config.user = user;
            config.password = password;
            config.autoBackup = autoBackup;
            config.autoBackupTime = autoBackupTime;
            config.autoLogin = autoLogin;
            config.rdiffCommand = rdiffCommand;
            FileOutputStream fout = new FileOutputStream(configFilename);
            ObjectOutputStream oos = new ObjectOutputStream(fout);
            oos.writeObject(config);
            oos.close();
            ui.status("Configuration saved.");
        }
        catch (Exception e) {
            e.printStackTrace();
        }
    }

    /*
     * ----------------- Private methods -----------------
     */

    /*
     * Loads the configuration from the file config.dat or calls initConfig() if
     * the file does not exist.
     */
    private void loadConfig() {
        File configFile = new File(configFilename);
        // Load configuration from file if it exists...
        if (configFile.isFile()) {
            try {
                FileInputStream fin = new FileInputStream(configFilename);
                ObjectInputStream ois = new ObjectInputStream(fin);
                config = (Config)ois.readObject();
                ois.close();

                targets = config.targets;
                ignores = config.ignores;
                host = config.host;
                port = config.port;
                hostname = config.hostname;
                user = config.user;
                password = config.password;
                autoBackup = config.autoBackup;
                autoBackupTime = config.autoBackupTime;
                autoLogin = config.autoLogin;
                rdiffCommand = config.rdiffCommand;
            }
            catch (Exception e) {
                e.printStackTrace();
            }
        }
        // ...else initialize new configuration
        else {
            initConfig();
        }
    }

    /*
     * Initialize a new configuration
     */
    private void initConfig() {
        config = new Config();

        targets = new ArrayList<String>();
        ignores = new ArrayList<String>();
        host = "";
        port = 9798;
        hostname = "";
        user = "";
        password = "";
        autoBackup = false;
        autoBackupTime = "12:00";
        autoLogin = false;
        if (osName.startsWith("Windows")) {
            rdiffCommand = "c:\rdiff\rdiff-backup.bat";
        }
        else {
            rdiffCommand = "rdiff-backup";
        }

        // Default values
        targets.add(System.getProperty("user.home"));
        host = "localhost"; // TODO Should be set according to server address
        try {
            java.net.InetAddress localMachine = java.net.InetAddress
                    .getLocalHost();
            hostname = localMachine.getHostName();
        }
        catch (java.net.UnknownHostException e) {
        }
        user = System.getProperty("user.name");

        // Save the new configuration
        saveConfig();
    }

    /*
     * Generates rdiff-backup command line for backup of Unix systems
     */
    private ArrayList buildBackupCmdUnix() {
        ArrayList<String[]> cmds = new ArrayList<String[]>();
        ArrayList<String> cmd = new ArrayList<String>();
        cmd.add(rdiffCommand);
        cmd.add("-b");
        cmd.add("-v5");

        for (int i = 0; i < targets.size(); i++) {
            cmd.add("--include");
            cmd.add(targets.get(i));
        }

        for (int i = 0; i < ignores.size(); i++) {
            cmd.add("--exclude");
            cmd.add(ignores.get(i));
        }

        cmd.add("--exclude");
        cmd.add("/");
        cmd.add("/");
        cmd.add(user + "@" + host + "::backup/" + hostname);

        String[] type = {};

        cmds.add(cmd.toArray(type));

        return cmds;
    }

    /*
     * Generates rdiff-backup command line for backup of Windows systems
     */
    private ArrayList buildBackupCmdWin() {
        ArrayList<String[]> cmds = new ArrayList<String[]>();
        ArrayList<String> cmd = new ArrayList<String>();
        Hashtable<String, ArrayList<String>> drives =
	    new Hashtable<String, ArrayList<String>>();
        Hashtable<String, ArrayList<String>> ignoredrives =
	    new Hashtable<String, ArrayList<String>>();

        // Build list of backup targets for each drive
        for (int i = 0; i < targets.size(); i++) {
            String target = targets.get(i);
            String drive = target.substring(0, 1);
            if (!drives.containsKey(drive)) {
                ArrayList<String> t = new ArrayList<String>();
                t.add(target);
                drives.put(drive, t);
            }
            else {
                ArrayList<String> t = drives.get(drive);
                t.add(target);
            }
        }

        // Build list of backup ignores for each drive
        for (int i = 0; i < ignores.size(); i++) {
            String ignore = ignores.get(i);
            String drive = ignore.substring(0, 1);
            if (drives.containsKey(drive)) {
                if (!ignoredrives.containsKey(drive)) {
                    ArrayList<String> t = new ArrayList<String>();
                    t.add(ignore);
                    ignoredrives.put(drive, t);
                }
                else {
                    ArrayList<String> t = ignoredrives.get(drive);
                    t.add(ignore);
                }
            }
        }

        // Run through the different drive letters
        for (Enumeration e = drives.keys(); e.hasMoreElements();) {
            String key = ((String)e.nextElement()).toLowerCase();

            cmd.add(rdiffCommand);
            cmd.add("-b");
            cmd.add("-v5");

            ArrayList d = drives.get(key);

            // Include all the targets for current drive letter
            for (int i = 0; i < d.size(); i++) {
                cmd.add("--include");
                cmd.add((String)d.get(i));
            }

            // Exclude all (if any) ignores for current drive letter
            if (ignoredrives.containsKey(key)) {
                ArrayList id = ignoredrives.get(key);

                for (int i = 0; i < id.size(); i++) {
                    cmd.add("--exclude");
                    cmd.add((String)d.get(i));
                }
            }

            cmd.add("--exclude");
            cmd.add(key + ":\\");
            cmd.add(key + ":\\");
            cmd.add(user + "@" + host + "::backup/" + hostname + "/" + key
                    + "_drive_rdiff-backup");

            String[] type = {};
            cmds.add(cmd.toArray(type));
        }

        return cmds;
    }

    /*
     * Generates rdiff-backup command line for restore of Unix systems
     */
    private String[] buildRestoreCmdUnix(String timestamp, String target,
            Boolean force) {
        ArrayList<String> cmd = new ArrayList<String>();
        cmd.add(rdiffCommand);
        cmd.add("-v5");
        if (force) {
            cmd.add("--force");
        }
        cmd.add("-r");
        cmd.add(timestamp);
        cmd.add(user + "@" + host + "::backup/" + hostname + target);
        cmd.add(target);
        String[] type = {};
        return cmd.toArray(type);
    }

    /*
     * Generates rdiff-backup command line for restore of Windows systems
     */
    private String[] buildRestoreCmdWin(String timestamp, String target,
            Boolean force) {
        ArrayList<String> cmd = new ArrayList<String>();

        String drive = target.substring(0, 1);
        String from = target.substring(2).replace("\\", "/");

        cmd.add(rdiffCommand);
        cmd.add("-v5");
        if (force) {
            cmd.add("--force");
        }
        cmd.add("-r");
        cmd.add(timestamp);
        cmd.add(user + "@" + host + "::backup/" + hostname + "/" + drive
                + "_drive_rdiff-backup" + from);
        cmd.add(target);

        String[] type = {};
        return cmd.toArray(type);
    }

    /*
     * Returns a SHA-1 hash of the given string.
     */
    private String hash(String s) {
        MessageDigest alg = null;
        byte[] buffer;
        String result;

        try {
            alg = MessageDigest.getInstance("SHA-1");
        }
        catch (java.security.NoSuchAlgorithmException e) {
            return null;
        }

        buffer = s.getBytes();
        alg.reset();
        alg.update(buffer);
        byte[] digest = alg.digest();

        StringBuffer hexString = new StringBuffer();
        for (int i = 0; i < digest.length; i++) {
            hexString.append(Integer.toHexString(0xFF & digest[i]));
        }
        result = hexString.toString();
        // System.out.println(s + " = " + result);
        return result;
    }

    /*
     * Sends dummy XML-RPC call to server to keep connection alive.
     */
    private void keepAlive() {
        Vector<String> query = new Vector<String>();
        query.add(hash(hash(password) + nextChallenge));
        query.add(user);
        conn.send("keepAlive", query);
    }
}
