package backup.client;

import java.io.*;
import java.util.*;
import java.text.SimpleDateFormat;

/**
 * Command line interface for <code>BackupClient</code>, implementing
 * <code>UserInterface</code>.
 * 
 * @author <a href="mailto:mauset@idi.ntnu.no">Tore Mauset</a>
 */
public class Cli extends Thread implements UserInterface {

    BackupClient bc;
    boolean running = true;
    InputStream inputstream;
    InputStreamReader inputstreamreader;
    BufferedReader bufferedreader;

    /**
     * Constructs a new Cli.
     * 
     * @param bc parent BackupClient
     */
    public Cli(BackupClient bc) {
        this.bc = bc;

        inputstream = System.in;
        inputstreamreader = new InputStreamReader(inputstream);
        bufferedreader = new BufferedReader(inputstreamreader);

        start();
    }

    /**
     * Start a separate thread for keeping track of keyboard input.
     */
    public void run() {
        try {
            String string = null;

            System.out.print("Not logged in > ");
            while (running) {
                string = bufferedreader.readLine();
                if (!string.equals("")) {
                    in(string);
                }
                String user = "Not logged in";
                String hostname = "";
                if (bc.isLoggedIn()) {
                    user = bc.getUser();
                    hostname = "@" + bc.getHostname();
                }
                if (running) {
                    System.out.print(user + hostname + " > ");
                }
            }
        }
        catch (Exception e) {
            e.printStackTrace();
        }
    }

    /*
     * (non-Javadoc)
     * 
     * @see backup.client.UserInterface#status(java.lang.String)
     */
    public void status(String msg) {
        Date today = new Date();
        SimpleDateFormat formatter = new SimpleDateFormat("HH:mm:ss");
        String time = formatter.format(today);
        System.out.println(time + " " + msg);
    }

    /*
     * (non-Javadoc)
     * 
     * @see backup.client.UserInterface#warn(java.lang.String)
     */
    public void warn(String msg) {
        Date today = new Date();
        SimpleDateFormat formatter = new SimpleDateFormat("HH:mm:ss");
        String time = formatter.format(today);
        System.out.println(time + " [!] " + msg);
    }

    /*
     * (non-Javadoc)
     * 
     * @see backup.client.UserInterface#processingFile(java.lang.String)
     */
    public void processingFile(String filename) {
        Date today = new Date();
        SimpleDateFormat formatter = new SimpleDateFormat("HH:mm:ss");
        String time = formatter.format(today);
        System.out.println(time + " Processing: " + filename);
    }

    /*
     * Processes one line of user input read from the keyboard.
     */
    private void in(String msg) {
        String[] tmp = msg.split(" ", 2);
        String cmd = tmp[0];
        String allargs = "";
        String[] args = {};

        if (tmp.length > 1) {
            allargs = tmp[1];
            args = allargs.split(" ");
        }

        if (cmd.equals("exit") && args.length == 0) {
            bc.exit();
            running = false;
        }
        else if (cmd.equals("add") && args.length > 0) {
            bc.addTarget(allargs);
            listTargets();
        }
        else if (cmd.equals("remove") && args.length > 0) {
            bc.removeTarget(allargs);
            listTargets();
        }
        else if (cmd.equals("ignore") && args.length > 0) {
            bc.addIgnore(allargs);
            listIgnores();
        }
        else if (cmd.equals("unignore") && args.length > 0) {
            bc.removeIgnore(allargs);
            listIgnores();
        }
        else if (cmd.equals("list")) {
            if (args.length == 0) {
                listTargets();
                listIgnores();
            }
            else if (args[0].equals("targets")) {
                listTargets();
            }
            else if (args[0].equals("ignores")) {
                listIgnores();
            }
        }
        else if (cmd.equals("files")) {
            String search;
            if (args.length > 0) {
                search = allargs;
            }
            else {
                search = "";
            }
            listFiles(search);
        }
        else if (cmd.equals("set")) {
            if (args.length == 0) {
                status("user = " + bc.getUser());
                status("password = <hidden>");
                status("host = " + bc.getHost());
                status("port = " + bc.getPort());
                status("hostname = " + bc.getHostname());
                if (bc.getAutoBackup()) {
                    status("autobackup = true");
                }
                else {
                    status("autobackup = false");
                }
                status("autobackuptime = " + bc.getAutoBackupTime());
                if (bc.getAutoLogin()) {
                    status("autologin = true");
                }
                else {
                    status("autologin = false");
                }
                status("rdiffcommand = " + bc.getRdiffCommand());
            }
            else if (args.length == 1) {
                if (args[0].equals("user")) {
                    status("user = " + bc.getUser());
                }
                else if (args[0].equals("password")) {
                    status("password = <hidden>");
                }
                else if (args[0].equals("host")) {
                    status("host = " + bc.getHost());
                }
                else if (args[0].equals("port")) {
                    status("port = " + bc.getPort());
                }
                else if (args[0].equals("hostname")) {
                    status("hostname = " + bc.getHostname());
                }
                else if (args[0].equals("autobackup")) {
                    if (bc.getAutoBackup()) {
                        status("autobackup = true");
                    }
                    else {
                        status("autobackup = false");
                    }
                }
                else if (args[0].equals("autobackuptime")) {
                    status("autobackuptime = " + bc.getAutoBackupTime());
                }
                else if (args[0].equals("autologin")) {
                    if (bc.getAutoLogin()) {
                        status("autologin = true");
                    }
                    else {
                        status("autologin = false");
                    }
                }
                else if (args[0].equals("rdiffcommand")) {
                    status("rdiffcommand = " + bc.getRdiffCommand());
                }

            }
            else {
                if (args[0].equals("user")) {
                    String user = allargs.split(" ", 2)[1];
                    bc.setUser(user);
                    status("user = " + user);
                }
                else if (args[0].equals("password")) {
                    String password = allargs.split(" ", 2)[1];
                    bc.setPassword(password);
                    status("password = " + password);
                }
                else if (args[0].equals("host")) {
                    String host = allargs.split(" ", 2)[1];
                    bc.setHost(host);
                    status("host = " + host);
                }
                else if (args[0].equals("port")) {
                    String port = allargs.split(" ", 2)[1];
                    bc.setPort(port);
                    status("port = " + port);
                }
                else if (args[0].equals("hostname")) {
                    String hostname = allargs.split(" ", 2)[1];
                    bc.setHostname(hostname);
                    status("hostname = " + hostname);
                }
                else if (args[0].equals("autobackup")) {
                    String autobackup = allargs.split(" ", 2)[1];
                    if (autobackup.equalsIgnoreCase("true")) {
                        bc.setAutoBackup(true);
                        status("autobackup = true");
                    }
                    else if (autobackup.equalsIgnoreCase("false")) {
                        bc.setAutoBackup(false);
                        status("autobackup = false");
                    }

                }
                else if (args[0].equals("autobackuptime")) {
                    // TODO Check that argument is a valid 24h timestamp
                    String autobackuptime = allargs.split(" ", 2)[1];
                    bc.setAutoBackupTime(autobackuptime);
                    status("autobackuptime = " + autobackuptime);
                }
                else if (args[0].equals("autologin")) {
                    String autologin = allargs.split(" ", 2)[1];
                    if (autologin.equalsIgnoreCase("true")) {
                        bc.setAutoLogin(true);
                        status("autologin = true");
                    }
                    else if (autologin.equalsIgnoreCase("false")) {
                        bc.setAutoLogin(false);
                        status("autologin = false");
                    }

                }
                else if (args[0].equals("rdiffcommand") && args.length > 0) {
                    String rdiffCommand = allargs.split(" ", 2)[1];
                    bc.setRdiffCommand(rdiffCommand);
                    status("rdiffcommand = " + rdiffCommand);
                }
            }
        }
        else if (cmd.equals("save") && args.length == 0) {
            bc.saveConfig();
        }
        else if (cmd.equals("backup")) {
            bc.startBackup();
        }
        else if (cmd.equals("restore") && args.length > 0) {
            String timestamp;
            String target;
            Boolean force;
            if (args[0].equals("-f")) {
                status("force");
                force = true;
                if (args.length > 3 && args[1].equals("-t")) {
                    timestamp = args[2];
                    target = allargs.split(" ", 4)[3];
                }
                else {
                    timestamp = "now";
                    target = allargs.split(" ", 2)[1];
                }
            }
            else {
                force = false;
                if (args.length > 2 && args[0].equals("-t")) {
                    timestamp = args[1];
                    target = allargs.split(" ", 3)[2];
                }
                else {
                    timestamp = "now";
                    target = allargs;
                }
            }
            if (target.contains("*")) {
                warn("Wildcards not allowed.");
            }
            else {
                bc.startRestore(timestamp, target, force);
            }
        }
        else if (cmd.equals("login") && args.length == 0) {
            bc.login();
        }
        else if (cmd.equals("logout") && args.length == 0) {
            bc.logout();
        }
        else if (cmd.equals("help")) {
            usage();
        }

    }

    /*
     * Prints help.
     */
    private void usage() {
        status("login                           login on backup server");
        status("add <target>                    add a new target for backup");
        status("remove <target>                 remove a target for backup");
        status("ignore <target>                 add an exception from backup");
        status("unignore <target>               remove an exception from backup");
        status("list [targets|ignores]          lists targets and/or exceptions");
        status("set [option [value]]            views or sets configuration");
        status("save                            saves current configuration");
        status("backup                          start manual backup");
        status("restore [-f] [-t ts] <target>   start restore");
        status("                                -f specifies force overwrite, USE");
        status("                                WITH CAUTION!!");
        status("                                ts is timestamp from rdiff-backup");
        status("                                default ts is \"now\"");
        status("files [search term]             list files available for restore");
        status("                                search term may contain *");
        status("exit                            exit client");
        status("help                            this text");
    }

    /*
     * Prints a list of current targets.
     */
    private void listTargets() {
        ArrayList t = bc.listTargets();

        if (t != null) {
            status("Targets:");
            for (int i = 0; i < t.size(); i++) {
                status((String)t.get(i));
            }
        }

    }

    /*
     * Prints a list of current ignores.
     */
    private void listIgnores() {
        ArrayList t = bc.listIgnores();

        if (t != null) {
            status("Ignores:");
            for (int i = 0; i < t.size(); i++) {
                status((String)t.get(i));
            }
        }
    }

    /*
     * Prints a list of files matching search term and their available
     * timestamps
     */
    private void listFiles(String search) {
        status("Searching...");
        Hashtable result = bc.listFiles(search);
        if (result == null) {
            return;
        }
        else if (result.isEmpty()) {
            warn("No files found matching your search.");
        }
        else {
            ArrayList<String> tmplist = new ArrayList<String>();
            Hashtable<String, Vector<String>> filelist =
		new Hashtable<String, Vector<String>>();

            // Run through hashtable returned from server
            for (Enumeration e = result.keys(); e.hasMoreElements();) {
                String k = (String)e.nextElement();
                String entry = (String)result.get(k);
                String timestamp = entry.substring(0, 25);
                String filename = entry.substring(26);
                // If we've seen this filename before, just add another
                // timestamp
                if (tmplist.contains(filename)) {
                    ((Vector<String>)filelist.get(filename)).add(timestamp);
                }
                // Else add it to the list of filenames, and add a new timestamp
                else {
                    Vector<String> v = new Vector<String>();
                    v.add(timestamp);
                    filelist.put(filename, v);
                    tmplist.add(filename);

                }
            }

            // Create a alphabetized list of filenames
            String[] filenamelist;
            String[] a = {};
            filenamelist = (tmplist.toArray(a));
            Arrays.sort(filenamelist);

            status("Files matching your search:");

            // Run through the files alphabetically and list their names and
            // timestamps
            for (int i = 0; i < filenamelist.length; i++) {
                String filename = filenamelist[i];
                String item = filename + " - ";
                Vector<String> v = filelist.get(filename);
                item += (String)v.remove(0);
                while (!v.isEmpty()) {
                    item += ", " + (String)v.remove(0);
                }
                status(item);
            }
        }
    }

}
