package backup.server;

import java.io.*;
import java.security.MessageDigest;
import java.text.SimpleDateFormat;
import java.util.*;
import java.util.Date;
import java.util.zip.*;
import java.sql.*;

import org.apache.xmlrpc.*;

/**
 * Main server engine.
 * 
 * @author <a href="mailto:mauset@idi.ntnu.no">Tore Mauset</a>
 */
public class BackupServer extends Thread {
    private WebServer server;
    private Boolean running;
    private Boolean verbose = false;
    private Cli ui;
    private ArrayList<String> allUsers;
    private ArrayList<String> currentUsers;
    private Hashtable<String, String> currentChallenges;
    private Hashtable<String, Calendar> lastSeen;
    private Random r;
    private Statement stmt;
    private Connection con;
    private int port;
    private String dbUrl;

    private final String dbHost = "localhost";
    private final String dbPort = "3306";
    private final String dbName = "BackupServer";
    private final String dbUser = "BackupServer";
    private final String dbPass = "testrun";

    /**
     * Constructs a new BackupServer.
     * 
     * @param args command line arguments
     */
    public BackupServer(String[] args) {
        port = 9798;

        ui = new Cli(this);
        allUsers = new ArrayList<String>();
        currentUsers = new ArrayList<String>();
        currentChallenges = new Hashtable<String, String>();
        lastSeen = new Hashtable<String, Calendar>();

        r = new Random();

        // Parse command line parameters
        if (args.length > 0) {
            for (int i = 0; i < args.length; i++) {
                String arg = args[i];
                if (arg.equals("-v")) {
                    verbose = true;
                }
                else if (arg.equals("-p")) {
                    if (args.length > i + 1) {
                        try {
                            port = Integer.parseInt(args[i + 1]);
                        }
                        catch (NumberFormatException e) {
                            ui.status("Invalid port number.");
                            System.exit(1);
                        }
                        i++;
                    }
                    else {
                        ui.status("Please specify port number.");
                        System.exit(1);
                    }
                }
                else {
                    ui.status("Invalid option: " + arg);
                    System.exit(1);
                }
            }
        }

        loadUsers();

        // Start the XML-RPC server
        server = new WebServer(port);
        server.addHandler("BackupServer", new ConnectionListener(this));
        server.start();
        if (verbose) {
            ui.status("Listening for connections on port " + port);
        }

        ui.start();

        // Register the JDBC driver for MySQL
        try {
            Class.forName("com.mysql.jdbc.Driver");
            dbUrl = "jdbc:mysql://" + dbHost + ":" + dbPort + "/" + dbName;
        }
        catch (Exception e) {
            e.printStackTrace();
            System.exit(1);
        }

        running = true;
    }

    /**
     * Main.
     * 
     * @param args command line arguments
     */
    public static void main(String[] args) {
        BackupServer bs = new BackupServer(args);
        bs.run();

    }

    /**
     * Start a separate thread for keeping track of tasks performed at regular
     * intervals.
     */
    public void run() {
        SimpleDateFormat formatter = new SimpleDateFormat("HH:mm");
        String prev = "";
        int mins = 0;

        try {
            while (running) {
                Thread.sleep(1000);
                String now = formatter.format(new Date());
                if (!now.equals(prev)) {
                    prev = now;
                    mins++;

                    // Check for lost connections every minutes
                    checkConnections();

                    // Check for aborted backups
                    if (mins % 30 == 0) {
                        checkBackups();
                    }

                    // Reset counter every hour
                    if (mins == 60) {
                        mins = 0;
                    }
                }
            }
        }
        catch (InterruptedException e) {
            e.printStackTrace();
        }
    }

    /**
     * Shut down server.
     */
    public void exit() {
        ui.warn("Server shutting down.");
        server.shutdown();
        running = false;
    }

    /**
     * Generates a new password challenge.
     * 
     * @param user user requesting challenge
     * @return challenge
     */
    public String getChallenge(String user) {
        // Generate random seed for password challenge
        String challenge = Long.toString(Math.abs(r.nextLong()), 36);

        String tmp = getPWHash(user) + challenge;
        currentChallenges.put(user, hash(tmp));

        return challenge;
    }

    /**
     * Check a response against the issued challenge.
     * 
     * @param response response from the client
     * @param user user sending the response
     * @return true or false for success
     */
    public Boolean checkResponse(String response, String user) {
        if (!currentChallenges.containsKey(user)) {
            return false;
        }

        if (response.equals(currentChallenges.get(user))) {
            // Any successful response means we can update the clients
            // last seen record.
            try {
                lastSeen.put(user, Calendar.getInstance());
            }
            catch (Exception e) {
                e.printStackTrace();
            }
            return true;
        }
        else {
            return false;
        }
    }

    /**
     * Start login procedure for a user.
     * 
     * @param user username trying to log in
     * @return challenge for client
     */
    public String startLogin(String user) {
        if (verbose) {
            ui.status("User '" + user + "' trying to log in.");
        }
        return getChallenge(user);
    }

    /**
     * Conclude login procedure for a user.
     * 
     * @param user username trying to log in
     * @param response response to challenge
     * @return success
     */
    public Boolean finishLogin(String user, String response) {
        // Is this a valid user?
        if (!allUsers.contains(user)) {
            if (verbose) {
                ui.status("User '" + user + "' does not exist. Login failed.");
            }
            return false;
        }

        // Is the user already logged in?
        if (currentUsers.contains(user)) {
            if (verbose) {
                ui.status("User '" + user
                        + "' is already logged in. Login failed.");
            }
            return false;
        }

        // Is the response correct?
        if (checkResponse(response, user)) {
            currentUsers.add(user);
            currentChallenges.remove(user);
            if (verbose) {
                ui.status("User '" + user + "' logged in.");
            }
            return true;
        }
        else {
            currentChallenges.remove(user);
            if (verbose) {
                ui.status("User '" + user + "' failed login.");
            }
            return false;
        }
    }

    /**
     * Log out out a user from the server.
     * 
     * @param user username
     */
    public void logout(String user) {
        currentUsers.remove(user);
        if (verbose) {
            ui.status("User '" + user + "' logged out.");
        }
    }

    /**
     * Get users currently logged in
     * 
     * @return users currently logged in
     */
    public ArrayList getUsers() {
        return currentUsers;
    }

    /**
     * Checks for incomplete backups in the specified user's home directory on
     * the server. Fixes problems according to
     * http://www.nongnu.org/rdiff-backup/FAQ.html#regress_failure
     * 
     * @param user username
     */
    public void checkBackup(String user) {
        ArrayList<String[]> list = findMetaDirs(user);

        // Filename filter for current_mirror*.data
        FilenameFilter currentmirrorFilter = new FilenameFilter() {
            public boolean accept(File dir, String name) {
                return (name.startsWith("current_mirror") && name
                        .endsWith(".data"));
            }
        };

        // Filename filter for mirror_metadata*.snapshot.gz
        FilenameFilter mirrormetadataFilter = new FilenameFilter() {
            public boolean accept(File dir, String name) {
                return (name.startsWith("mirror_metadata") && name
                        .endsWith(".snapshot.gz"));
            }
        };

        // Run through all the user's meta dirs
        for (int i = 0; i < list.size(); i++) {
            File dir = new File(list.get(i)[1]);

            if (verbose) {
                ui.status("Checking " + list.get(i)[1]);
            }

            // Find and delete duplicate current_mirror.*.data files...
            String[] currentmirrorFiles = dir.list(currentmirrorFilter);
            if (currentmirrorFiles == null) {
                return;
            }
            else if (currentmirrorFiles.length > 1) {
                System.out.println("kom hit");
                Arrays.sort(currentmirrorFiles);
                File deleteMe = new File(dir.toString() + "/"
                        + currentmirrorFiles[0]);
                if (verbose) {
                    ui.status("Deleted file " + deleteMe);
                }
                deleteMe.delete();

                // ...and rename corrupted mirror_metadata.*.snapshot.gz
                String[] mirrormetadataFiles = dir.list(mirrormetadataFilter);

                Arrays.sort(mirrormetadataFiles);
                File renameMe = new File(dir.toString() + "/"
                        + mirrormetadataFiles[mirrormetadataFiles.length - 1]);
                File newName = new File(renameMe.toString().replaceAll(
                        "mirror_metadata", "aborted-metadata"));
                if (verbose) {
                    ui.status("Renamed file " + renameMe + " to " + newName);
                }
                renameMe.renameTo(newName);
            }
        }
    }

    /**
     * Reads file statistics from rdiff-backup logs and inserts records for new
     * and updated files into the database.
     * 
     * @param user username
     */
    public void updateDB(String user) {
        int fileCount = 0;
        ArrayList<String[]> metaDirs = findMetaDirs(user);

        // Filename filter for file_statistics*.data.gz
        FilenameFilter filter = new FilenameFilter() {
            public boolean accept(File dir, String name) {
                return (name.startsWith("file_statistics") && name
                        .endsWith(".data.gz"));
            }
        };

        try {
            con = DriverManager.getConnection(dbUrl, dbUser, dbPass);

            for (int i = 0; i < metaDirs.size(); i++) {
                String root = metaDirs.get(i)[0];
                File dir = new File(metaDirs.get(i)[1]);
                String[] statFiles = dir.list(filter);
                Arrays.sort(statFiles);

                // Only process the newest statistics file
                String gzipFilename = statFiles[statFiles.length - 1];
                String time = gzipFilename.substring(16, 41);

                GZIPInputStream gzipInputStream;
                try {
                    gzipInputStream = new GZIPInputStream(new FileInputStream(
                            dir.toString() + "/" + gzipFilename));
                    BufferedReader in = new BufferedReader(
                            new InputStreamReader(gzipInputStream));
                    String line = null;

                    // Read each line from the log file
                    while ((line = in.readLine()) != null) {
                        int tmp;
                        tmp = line.lastIndexOf(" ") + 1;
                        String incrementSize = line.substring(tmp, line.length());
                        line = line.substring(0, tmp - 1);
                        tmp = line.lastIndexOf(" ") + 1;
                        String mirrorSize = line.substring(tmp, line.length());
                        line = line.substring(0, tmp - 1);
                        tmp = line.lastIndexOf(" ") + 1;
                        String sourceSize = line.substring(tmp, line.length());
                        line = line.substring(0, tmp - 1);
                        tmp = line.lastIndexOf(" ") + 1;
                        String changed = line.substring(tmp, line.length());
                        String filename = line.substring(0, tmp - 1);

                        // Changed or added. Not deleted.
                        if (changed.equals("1") && !sourceSize.equals("NA")
                                && !filename.equals(".")) {
                            insertDB(root + filename, user, time);
                            fileCount++;
                        }
                    }
                }
                catch (FileNotFoundException e) {
                    e.printStackTrace();
                }
                catch (IOException e) {
                    e.printStackTrace();
                }

            }
            con.close();
        }
        catch (SQLException e1) {
            // TODO Auto-generated catch block
            e1.printStackTrace();
        }

        ui.status("Updated/added " + String.valueOf(fileCount)
                + " files/directories for user '" + user + "'.");
    }

    /**
     * Returns a list of a user's files matching the specified search term, or
     * all files if search term is blank.
     * 
     * @param user username
     * @param search search term
     * @return list of files with timestamps
     */
    public Hashtable<String, String> getFiles(String user, String search) {
        Hashtable<String, String> result = new Hashtable<String, String>();
        try {
            con = DriverManager.getConnection(dbUrl, dbUser, dbPass);
            ResultSet rs;
            stmt = con.createStatement();
            String query = "SELECT * FROM `files` WHERE `user` LIKE '" + user
                    + "'";

            if (!search.equals("")) {
                query += " AND `filename` LIKE '" + search + "'";
            }
            query += ";";

            rs = stmt.executeQuery(query);

            while (rs.next()) {
                String id = String.valueOf(rs.getInt("id"));
                String timestamp = rs.getString("timestamp");
                String filename = rs.getString("filename");

                result.put(id, timestamp + "_" + filename);
            }
            con.close();
        }
        catch (SQLException e) {
            e.printStackTrace();
        }

        return result;
    }

    /*
     * ----------------- Private methods -----------------
     */

    /*
     * Returns an ArrayList of all the meta data directories and the root
     * directory they correspond to on the client.
     */
    private ArrayList<String[]> findMetaDirs(String user) {
        File root = new File("/home/" + user + "/backup");
        ArrayList<String[]> metadirs = new ArrayList<String[]>();

        if (!root.isDirectory()) {
            return metadirs;
        }

        String[] dirs = root.list();

        for (String dirname : dirs) {
            File tmp = new File(root.toString() + "/" + dirname
                    + "/rdiff-backup-data/");
            if (tmp.isDirectory()) {
                String[] mdir = { "/", tmp.toString() };
                metadirs.add(mdir);
            }

            // Need to check deeper in case of Windows backup
            File dir = new File(root.toString() + "/" + dirname);
            String[] subdirs = dir.list();
            for (String subdirname : subdirs) {
                if (subdirname.endsWith("_drive_rdiff-backup")) {
                    File subdir = new File(root.toString() + dirname + "/"
                            + subdirname + "/rdiff-backup-data/");
                    if (subdir.isDirectory()) {
                        String[] mdir = {
                                subdirname.substring(0, 2).toUpperCase()
                                        + ":\\", tmp.toString() };
                        metadirs.add(mdir);
                    }
                }
            }
        }

        return metadirs;
    }

    /*
     * Returns a SHA-1 hash of the given string.
     */
    private String hash(String s) {
        MessageDigest alg = null;
        byte[] buffer;
        String result;

        try {
            alg = MessageDigest.getInstance("SHA-1");
        }
        catch (java.security.NoSuchAlgorithmException e) {
            return null;
        }

        buffer = s.getBytes();
        alg.reset();
        alg.update(buffer);
        byte[] digest = alg.digest();

        StringBuffer hexString = new StringBuffer();
        for (int i = 0; i < digest.length; i++) {
            hexString.append(Integer.toHexString(0xFF & digest[i]));
        }
        result = hexString.toString();
        // System.out.println(s + " = " + result);
        return result;
    }

    /*
     * Checks whether all logged in clients have reported to the server during
     * the last three minutes.
     */
    private void checkConnections() {
        for (int i = 0; i < currentUsers.size(); i++) {
            String user = (String) currentUsers.get(i);
            Long then = ((Calendar) lastSeen.get(user)).getTimeInMillis() / 1000;
            Long now = Calendar.getInstance().getTimeInMillis() / 1000;
            Long diff = now - then;

            if (diff > 180) {
                if (verbose) {
                    ui.status("User '" + user + "' timed out. (Last seen "
                            + diff.toString() + " seconds ago)");
                }
                logout(user);
            }
        }
    }

    /*
     * Run through all the users, calling checkBackup on each one.
     */
    private void checkBackups() {
        for (int i = 0; i < allUsers.size(); i++) {
            checkBackup(allUsers.get(i));
        }
    }

    /*
     * Inserts an entry into the file database.
     */
    private void insertDB(String filename, String user, String timestamp) {
        String date = timestamp.substring(0, 16).replace("T", " ");
        try {
            stmt = con.createStatement();
            String query = "INSERT INTO `files` (`filename`, `user`, `date`, `timestamp`) "
                    + "VALUES ('"
                    + filename
                    + "', '"
                    + user
                    + "', '"
                    + date
                    + "', '" + timestamp + "');";
            stmt.executeUpdate(query);
        }
        catch (SQLException e) {
            e.printStackTrace();
        }
    }

    /*
     * Returns a SHA-1 hash of the user's password. Should be customized
     * according to the user/password backend.
     */
    private String getPWHash(String user) {
        if (!allUsers.contains(user)) {
            return hash("notfound");
        }
        return hash("test");
        // TODO Return real password hashes
    }

    /*
     * Load a list of all registered users. Should be customized according to
     * the user/password backend.
     */
    private void loadUsers() {
        allUsers.add("mauset");
        allUsers.add("test");
        // TODO Real loading
    }
}
