package backup.server;

import java.util.*;

/**
 * Handler for XML-RPC calls from <code>BackupClient</code>.
 * 
 * @author <a href="mailto:mauset@idi.ntnu.no">Tore Mauset</a>
 */
public class ConnectionListener {
    BackupServer bs;

    /**
     * Constructs a new ConnectionListener.
     * 
     * @param bs parent <code>BackupServer</code>
     */
    public ConnectionListener(BackupServer bs) {
        this.bs = bs;
    }

    /**
     * Start login procedure for specified user.
     * 
     * @param user username
     * @return hashtable containing new password challenge
     */
    public Hashtable startLogin(String user) {
        String challenge = bs.startLogin(user);
        Hashtable<String, String> result = new Hashtable<String, String>();
        result.put("challenge", challenge);
        return result;
    }

    /**
     * Finsh login procedure for specified user.
     * 
     * @param response response to previous password challenge
     * @param user username
     * @return hashtable containing success and new password challenge, or
     *         failure.
     */
    public Hashtable finishLogin(String response, String user) {
        Hashtable<String, String> result = new Hashtable<String, String>();
        Boolean success = bs.finishLogin(user, response);
        if (success) {
            String challenge = bs.getChallenge(user);
            result.put("challenge", challenge);
        }
        result.put("success", success.toString());
        return result;
    }

    /**
     * Log out specified user from server.
     * 
     * @param response response to previous password challenge
     * @param user username
     * @return empty hashtable
     */
    public Hashtable logout(String response, String user) {
        if (bs.checkResponse(response, user)) {
            bs.logout(user);
        }
        return new Hashtable();
    }

    /**
     * Dummy call to update BackupServer's Last Seen record for the specified
     * user.
     * 
     * @param response response to previous password challenge
     * @param user username
     * @return hashtable containing new password challenge
     */
    public Hashtable keepAlive(String response, String user) {
        // checkResponse will update the last seen record if the challenge
        // response is correct
        bs.checkResponse(response, user);
        String challenge = bs.getChallenge(user);
        Hashtable<String, String> result = new Hashtable<String, String>();
        result.put("challenge", challenge);
        return result;
    }

    /**
     * Tells BackupServer to check meta data of specified user.
     * 
     * @param response response to previous password challenge
     * @param user username
     * @return hashtable containing new password challenge
     */
    public Hashtable checkBackup(String response, String user) {
        Hashtable<String, String> result = new Hashtable<String, String>();
        if (bs.checkResponse(response, user)) {
            bs.checkBackup(user);
            String challenge = bs.getChallenge(user);
            result.put("challenge", challenge);
        }
        return result;
    }

    /**
     * Tells BackupServer to update database for specified user.
     * 
     * @param response response to previous password challenge
     * @param user username
     * @return hashtable containing new password challenge
     */
    public Hashtable updateDB(String response, String user) {
        Hashtable<String, String> result = new Hashtable<String, String>();
        if (bs.checkResponse(response, user)) {
            bs.updateDB(user);
            String challenge = bs.getChallenge(user);
            result.put("challenge", challenge);
        }
        return result;
    }

    /**
     * Requests a list of files matching the search term from BackupServer.
     * 
     * @param response response to previous password challenge
     * @param user username
     * @param search search term
     * @return hashtable containing list of files and new password challenge
     */
    public Hashtable listFiles(String response, String user, String search) {
        Hashtable<String, String> result = new Hashtable<String, String>();
        if (bs.checkResponse(response, user)) {
            result = bs.getFiles(user, search);
            String challenge = bs.getChallenge(user);
            result.put("challenge", challenge);
        }
        return result;
    }

}
